from lxml.etree import XPath, ElementBase
from lxml.html import fromstring, XHTML_NAMESPACE
from lxml.html import _forms_xpath, _options_xpath, _nons, _transform_result
from lxml.html import defs
import copy

try:
    basestring
except NameError:
    # Python 3
    basestring = str

__all__ = ['FormNotFound', 'fill_form', 'fill_form_html',
           'insert_errors', 'insert_errors_html',
           'DefaultErrorCreator']

class FormNotFound(LookupError):
    """
    Raised when no form can be found
    """

_form_name_xpath = XPath('descendant-or-self::form[name=$name]|descendant-or-self::x:form[name=$name]', namespaces={'x':XHTML_NAMESPACE})
_input_xpath = XPath('|'.join(['descendant-or-self::'+_tag for _tag in ('input','select','textarea','x:input','x:select','x:textarea')]),
                               namespaces={'x':XHTML_NAMESPACE})
_label_for_xpath = XPath('//label[@for=$for_id]|//x:label[@for=$for_id]',
                               namespaces={'x':XHTML_NAMESPACE})
_name_xpath = XPath('descendant-or-self::*[@name=$name]')

def fill_form(
    el,
    values,
    form_id=None,
    form_index=None,
    ):
    el = _find_form(el, form_id=form_id, form_index=form_index)
    _fill_form(el, values)

def fill_form_html(html, values, form_id=None, form_index=None):
    result_type = type(html)
    if isinstance(html, basestring):
        doc = fromstring(html)
    else:
        doc = copy.deepcopy(html)
    fill_form(doc, values, form_id=form_id, form_index=form_index)
    return _transform_result(result_type, doc)

def _fill_form(el, values):
    counts = {}
    if hasattr(values, 'mixed'):
        # For Paste request parameters
        values = values.mixed()
    inputs = _input_xpath(el)
    for input in inputs:
        name = input.get('name')
        if not name:
            continue
        if _takes_multiple(input):
            value = values.get(name, [])
            if not isinstance(value, (list, tuple)):
                value = [value]
            _fill_multiple(input, value)
        elif name not in values:
            continue
        else:
            index = counts.get(name, 0)
            counts[name] = index + 1
            value = values[name]
            if isinstance(value, (list, tuple)):
                try:
                    value = value[index]
                except IndexError:
                    continue
            elif index > 0:
                continue
            _fill_single(input, value)

def _takes_multiple(input):
    if _nons(input.tag) == 'select' and input.get('multiple'):
        # FIXME: multiple="0"?
        return True
    type = input.get('type', '').lower()
    if type in ('radio', 'checkbox'):
        return True
    return False

def _fill_multiple(input, value):
    type = input.get('type', '').lower()
    if type == 'checkbox':
        v = input.get('value')
        if v is None:
            if not value:
                result = False
            else:
                result = value[0]
                if isinstance(value, basestring):
                    # The only valid "on" value for an unnamed checkbox is 'on'
                    result = result == 'on'
            _check(input, result)
        else:
            _check(input, v in value)
    elif type == 'radio':
        v = input.get('value')
        _check(input, v in value)
    else:
        assert _nons(input.tag) == 'select'
        for option in _options_xpath(input):
            v = option.get('value')
            if v is None:
                # This seems to be the default, at least on IE
                # FIXME: but I'm not sure
                v = option.text_content()
            _select(option, v in value)

def _check(el, check):
    if check:
        el.set('checked', '')
    else:
        if 'checked' in el.attrib:
            del el.attrib['checked']

def _select(el, select):
    if select:
        el.set('selected', '')
    else:
        if 'selected' in el.attrib:
            del el.attrib['selected']

def _fill_single(input, value):
    if _nons(input.tag) == 'textarea':
        input.text = value
    else:
        input.set('value', value)

def _find_form(el, form_id=None, form_index=None):
    if form_id is None and form_index is None:
        forms = _forms_xpath(el)
        for form in forms:
            return form
        raise FormNotFound(
            "No forms in page")
    if form_id is not None:
        form = el.get_element_by_id(form_id)
        if form is not None:
            return form
        forms = _form_name_xpath(el, name=form_id)
        if forms:
            return forms[0]
        else:
            raise FormNotFound(
                "No form with the name or id of %r (forms: %s)"
                % (id, ', '.join(_find_form_ids(el))))               
    if form_index is not None:
        forms = _forms_xpath(el)
        try:
            return forms[form_index]
        except IndexError:
            raise FormNotFound(
                "There is no form with the index %r (%i forms found)"
                % (form_index, len(forms)))

def _find_form_ids(el):
    forms = _forms_xpath(el)
    if not forms:
        yield '(no forms)'
        return
    for index, form in enumerate(forms):
        if form.get('id'):
            if form.get('name'):
                yield '%s or %s' % (form.get('id'),
                                     form.get('name'))
            else:
                yield form.get('id')
        elif form.get('name'):
            yield form.get('name')
        else:
            yield '(unnamed form %s)' % index

############################################################
## Error filling
############################################################

class DefaultErrorCreator(object):
    insert_before = True
    block_inside = True
    error_container_tag = 'div'
    error_message_class = 'error-message'
    error_block_class = 'error-block'
    default_message = "Invalid"

    def __init__(self, **kw):
        for name, value in kw.items():
            if not hasattr(self, name):
                raise TypeError(
                    "Unexpected keyword argument: %s" % name)
            setattr(self, name, value)

    def __call__(self, el, is_block, message):
        error_el = el.makeelement(self.error_container_tag)
        if self.error_message_class:
            error_el.set('class', self.error_message_class)
        if is_block and self.error_block_class:
            error_el.set('class', error_el.get('class', '')+' '+self.error_block_class)
        if message is None or message == '':
            message = self.default_message
        if isinstance(message, ElementBase):
            error_el.append(message)
        else:
            assert isinstance(message, basestring), (
                "Bad message; should be a string or element: %r" % message)
            error_el.text = message or self.default_message
        if is_block and self.block_inside:
            if self.insert_before:
                error_el.tail = el.text
                el.text = None
                el.insert(0, error_el)
            else:
                el.append(error_el)
        else:
            parent = el.getparent()
            pos = parent.index(el)
            if self.insert_before:
                parent.insert(pos, error_el)
            else:
                error_el.tail = el.tail
                el.tail = None
                parent.insert(pos+1, error_el)

default_error_creator = DefaultErrorCreator()
    

def insert_errors(
    el,
    errors,
    form_id=None,
    form_index=None,
    error_class="error",
    error_creator=default_error_creator,
    ):
    el = _find_form(el, form_id=form_id, form_index=form_index)
    for name, error in errors.items():
        if error is None:
            continue
        for error_el, message in _find_elements_for_name(el, name, error):
            assert isinstance(message, (basestring, type(None), ElementBase)), (
                "Bad message: %r" % message)
            _insert_error(error_el, message, error_class, error_creator)

def insert_errors_html(html, values, **kw):
    result_type = type(html)
    if isinstance(html, basestring):
        doc = fromstring(html)
    else:
        doc = copy.deepcopy(html)
    insert_errors(doc, values, **kw)
    return _transform_result(result_type, doc)

def _insert_error(el, error, error_class, error_creator):
    if _nons(el.tag) in defs.empty_tags or _nons(el.tag) == 'textarea':
        is_block = False
    else:
        is_block = True
    if _nons(el.tag) != 'form' and error_class:
        _add_class(el, error_class)
    if el.get('id'):
        labels = _label_for_xpath(el, for_id=el.get('id'))
        if labels:
            for label in labels:
                _add_class(label, error_class)
    error_creator(el, is_block, error)

def _add_class(el, class_name):
    if el.get('class'):
        el.set('class', el.get('class')+' '+class_name)
    else:
        el.set('class', class_name)

def _find_elements_for_name(form, name, error):
    if name is None:
        # An error for the entire form
        yield form, error
        return
    if name.startswith('#'):
        # By id
        el = form.get_element_by_id(name[1:])
        if el is not None:
            yield el, error
        return
    els = _name_xpath(form, name=name)
    if not els:
        # FIXME: should this raise an exception?
        return
    if not isinstance(error, (list, tuple)):
        yield els[0], error
        return
    # FIXME: if error is longer than els, should it raise an error?
    for el, err in zip(els, error):
        if err is None:
            continue
        yield el, err



from base64 import b64encode

from ..packages.six import b

ACCEPT_ENCODING = 'gzip,deflate'


def make_headers(keep_alive=None, accept_encoding=None, user_agent=None,
                 basic_auth=None, proxy_basic_auth=None, disable_cache=None):
    """
    Shortcuts for generating request headers.

    :param keep_alive:
        If ``True``, adds 'connection: keep-alive' header.

    :param accept_encoding:
        Can be a boolean, list, or string.
        ``True`` translates to 'gzip,deflate'.
        List will get joined by comma.
        String will be used as provided.

    :param user_agent:
        String representing the user-agent you want, such as
        "python-urllib3/0.6"

    :param basic_auth:
        Colon-separated username:password string for 'authorization: basic ...'
        auth header.

    :param proxy_basic_auth:
        Colon-separated username:password string for 'proxy-authorization: basic ...'
        auth header.

    :param disable_cache:
        If ``True``, adds 'cache-control: no-cache' header.

    Example::

        >>> make_headers(keep_alive=True, user_agent="Batman/1.0")
        {'connection': 'keep-alive', 'user-agent': 'Batman/1.0'}
        >>> make_headers(accept_encoding=True)
        {'accept-encoding': 'gzip,deflate'}
    """
    headers = {}
    if accept_encoding:
        if isinstance(accept_encoding, str):
            pass
        elif isinstance(accept_encoding, list):
            accept_encoding = ','.join(accept_encoding)
        else:
            accept_encoding = ACCEPT_ENCODING
        headers['accept-encoding'] = accept_encoding

    if user_agent:
        headers['user-agent'] = user_agent

    if keep_alive:
        headers['connection'] = 'keep-alive'

    if basic_auth:
        headers['authorization'] = 'Basic ' + \
            b64encode(b(basic_auth)).decode('utf-8')

    if proxy_basic_auth:
        headers['proxy-authorization'] = 'Basic ' + \
            b64encode(b(proxy_basic_auth)).decode('utf-8')

    if disable_cache:
        headers['cache-control'] = 'no-cache'

    return headers



u"""
Fixer for os.getcwd() -> os.getcwdu().
Also warns about "from os import getcwd", suggesting the above form.
"""

from lib2to3 import fixer_base
from lib2to3.fixer_util import Name

class FixGetcwd(fixer_base.BaseFix):

    PATTERN = u"""
              power< 'os' trailer< dot='.' name='getcwd' > any* >
              |
              import_from< 'from' 'os' 'import' bad='getcwd' >
              """

    def transform(self, node, results):
        if u"name" in results:
            name = results[u"name"]
            name.replace(Name(u"getcwdu", prefix=name.prefix))
        elif u"bad" in results:
            # Can't convert to getcwdu and then expect to catch every use.
            self.cannot_convert(node, u"import os, use os.getcwd() instead.")
            return
        else:
            raise ValueError(u"For some reason, the pattern matcher failed.")



from production.box._OrderChangerProductionBox import OrderChangerProductionBox


class KeepProdStateBox(OrderChangerProductionBox):
    def __init__(self, name, sons, states, logger=None):
        super(KeepProdStateBox, self).__init__(name, sons, logger)
        self.allowed_states = {}
        for state in states:
            self.allowed_states[state] = state

    def change_order(self, order):
        order.tagged_lines = ""
        return order



# encoding: utf-8
from datetime import (
    datetime,
    timedelta
)

from systools import sys_log_exception
from typing import (
    List,
    Dict
)

from basereport import (
    SimpleReport,
    TableReport,
    ReportColumnDefinition,
    AlignTypes
)
from report import (
    Presenter,
    Part
)
from report.command import (
    NewLineCommand,
    AlignCommand,
    RepeatCommand,
    FontCommand
)
from reportsdata.tefcancellationreport.dto import TefCancellationGeneratorDto
from reportsdata.tefcancellationreport.dto import TefCancellationHeaderDto
from reportsdata.tefcancellationreport.dto import TefCancellationBodyDto
from utils.report_formatters import (
    format_line,
    remove_accents,
    format_price
)


class TefCancellationReportPresenter(Presenter):

    def __init__(
        self,
        columns  # type: int
    ):
        # type: (...) -> None

        self.total = 0
        self.columns = columns
        self.header_column_definition = [
            ReportColumnDefinition(
                width=16,
                after_fill_text=': ',
                fill_with_char='.'
            ),
            ReportColumnDefinition(width=columns - 16)
        ]
        self.header_table_column_definitions = [
            ReportColumnDefinition(
                width=7,
                after_text=' '
            ),
            ReportColumnDefinition(
                width=31
            )
        ]
        self.items_table_column_definitions = [
            ReportColumnDefinition(
                width=38
            )
        ]
        self.footer_table_column_definitions = [
            ReportColumnDefinition(
                width=29,
                after_text=' ',
                align=AlignTypes.RIGHT
            ),
            ReportColumnDefinition(
                width=9,
                align=AlignTypes.RIGHT
            )
        ]

    def present(
        self,
        dto     # type: TefCancellationGeneratorDto
    ):
        # type: (...) -> SimpleReport

        try:
            parts = [Part(commands=[FontCommand(FontCommand.Font.B)])]
            parts.extend(self._build_header(dto.header))
            parts.extend(self._build_table(dto.body))
            parts.extend([self._new_line(2)])
        except Exception as _:
            sys_log_exception("Error building Benefits Report")
            raise

        return SimpleReport(parts, self.columns)

    def _build_header(
        self,
        header  # type: TefCancellationHeaderDto
    ):
        # type: (...) -> List

        parts = [
            self._new_line(),
            self._print_text(
                text=header.title,
                align='center'
            ),
            self._new_line(2),
            self._get_store_name(header),
            self._get_date_time(),
            self._get_period(header),
            self._new_line(2)
        ]

        return parts

    def _build_table(
        self,
        body    # type: TefCancellationBodyDto
    ):
        # type: (...) -> List

        tenders_by_date = body.tenders_by_date
        if len(tenders_by_date) == 0:
            return self._build_nothing_found()

        data = []
        for date, tenders in sorted(tenders_by_date.items()):
            data.extend(self._build_table_header(date))
            data.extend(self._build_table_rows(
                items=tenders,
                card_brands=body.card_brands
            ))
            data.extend(self._build_table_footer())

        return data

    def _build_table_header(
        self,
        date
    ):
        # type: (...) -> List

        date = datetime.strptime(
            date,
            "%Y-%m-%d"
        ).strftime("%d/%m/%Y")

        header = [
            TableReport(
                [[
                    "$TEF_CANCELLATION_REPORT_DATE",
                    date
                ]],
                self.header_table_column_definitions
            ),
            self._repeat_char(
                "-",
                self.columns
            )
        ]

        return header

    def _build_table_rows(
        self,
        items,          # type: List
        card_brands     # type: Dict
    ):
        # type: (...) -> List

        rows = []
        self.total = 0
        for item in items:
            value = float(item.tender_detail["tenderAmount"])
            self.total += value
            date_time = datetime.strptime(
                item.timestamp,
                "%Y-%m-%dT%H:%M:%S.%f"
            )
            row_1 = [
                self._convert_to_local_time(date_time).strftime("%H:%M"),
                remove_accents(item.tender_descr),
                remove_accents(card_brands[int(item.tender_detail["Bandeira"])].description)
            ]
            row_2 = [
                self._get_card_number(item.tender_detail["CardNumber"], 6),
                item.tender_detail["NsuHost"],
                format_price(value)
            ]
            rows.extend([
                self._print_text(" ".join(row_1)),
                self._new_line(),
                self._print_text(" ".join(row_2)),
                self._new_line(),
                self._repeat_char(
                    "-",
                    self.columns
                )
            ])

        return rows

    def _build_nothing_found(self):
        # type: () -> List

        parts = [
            self._repeat_char(
                "-",
                self.columns
            ),
            self._new_line(),
            self._print_text("$NO_CANCELLATION_ERRORS", align='center'),
            self._new_line(2),
            self._repeat_char(
                "-",
                self.columns
            )
        ]
        return parts

    def _build_table_footer(self):
        # type: () -> List

        data = [
            TableReport(
                [[
                    "$TEF_CANCELLATION_REPORT_TOTAL",
                    format_price(self.total)
                ]],
                self.footer_table_column_definitions
            ),
            self._new_line()
        ]

        return data

    @staticmethod
    def _new_line(repeat=1):
        # type: (...) -> Part

        data = Part(
            None,
            [NewLineCommand() for i in range(0, repeat)]
        )

        return data

    @staticmethod
    def _repeat_char(char, columns):
        # type: (...) -> Part

        data = Part(
            None,
            [
                RepeatCommand(
                    columns,
                    char
                ),
                NewLineCommand()
            ]
        )

        return data

    @staticmethod
    def _print_text(text, align='left'):
        # type: (...) -> Part

        if type(text) == list:
            text = " ".join(text)

        if align == 'center':
            align_command = AlignCommand(AlignCommand.Alignment.center)
        elif align == 'right':
            align_command = AlignCommand(AlignCommand.Alignment.right)
        else:
            align_command = AlignCommand(AlignCommand.Alignment.left)

        data = Part(
            remove_accents(text),
            [align_command]
        )

        return data

    @staticmethod
    def _convert_to_local_time(
        date_time   # type: datetime
    ):
        # type: (...) -> datetime

        return date_time - timedelta(hours=3)

    @staticmethod
    def _get_card_number(card_number, digits):
        card_number = card_number[:digits].ljust(8, 'X')
        card_number = card_number[:4] + "." + card_number[4:]
        return card_number


    def _get_store_name(
        self,
        header  # type: TefCancellationHeaderDto
    ):
        # type: (...) -> List

        data = format_line(
            remove_accents("$STORE"),
            header.store_name,
            self.columns,
            self.header_column_definition
        )

        return data[0]

    def _get_date_time(self):
        # type: () -> List

        current_datetime = datetime.now().strftime('%d/%m/%y %H:%M:%S')
        data = format_line(
            remove_accents("$DATE_TIME"),
            current_datetime,
            self.columns,
            self.header_column_definition
        )

        return data[0]

    def _get_period(
        self,
        header  # type: TefCancellationHeaderDto
    ):
        # type: (...) -> List

        period = "{0} - {1}".format(
            header.start_date.strftime('%d/%m/%y'),
            header.end_date.strftime('%d/%m/%y')
        )
        data = format_line(
            remove_accents("$PERIOD"),
            period,
            self.columns,
            self.header_column_definition
        )

        return data[0]



import sys

if sys.version_info[0] < 3:
    import inspect2 as inspect
else:
    import inspect

from typing import Type, Optional, Dict, Any, Union, Tuple

from httpmessageprocessor import ApiRequest
from messagebus import JsonHttpResponseMessage, HttpResponseMessage
from messageprocessorutil import UuidMessageProcessor

from ._ApiRequestParser import ApiRequestParser
from ._ExceptionFormatter import ExceptionFormatter
from ._util import convert_to_json
from .exception import \
    InteractorWithoutExecuteMethod, \
    InteractorWithMoreThanOneParameterOnExecute, \
    ParameterSentToNoParameterInteractor


class AutoMappedMessageProcessor(UuidMessageProcessor):
    def __init__(self,
                 request_type,
                 interactor,
                 exception_map=None,
                 success_status=200,
                 extra_mapping=None,
                 event_name=None,
                 response_content_type=None):
        # type: (Optional[Type], Any, Optional[Dict[type, Union[int, Tuple[int, str]]]], int, dict, Optional[str], Optional[str]) -> None  # noqa
        super(AutoMappedMessageProcessor, self).__init__(event_name)
        self.parser = ApiRequestParser(request_type, extra_mapping)
        self.success_status = success_status
        self.response_content_type = response_content_type

        if not hasattr(interactor, "execute"):
            raise InteractorWithoutExecuteMethod()
        if len(inspect.signature(interactor.execute).parameters) > 1:
            raise InteractorWithMoreThanOneParameterOnExecute()
        self.interactor_has_parameter = False
        if len(inspect.signature(interactor.execute).parameters) == 1:
            self.interactor_has_parameter = True
        self.interactor = interactor

        self.exception_formatter = ExceptionFormatter(exception_map)
        self.exception_map = exception_map

    def parse_data(self, api_request, event_type):
        # type: (ApiRequest, Optional[str]) -> Any
        return self.parser.parse(api_request)

    def call_business(self, obj):
        if obj is not None:
            if not self.interactor_has_parameter:
                raise ParameterSentToNoParameterInteractor()

            return self.interactor.execute(obj)
        else:
            if self.interactor_has_parameter:
                return self.interactor.execute(None)
            else:
                return self.interactor.execute()

    def format_response(self, message_bus, message, event, input_obj, result):
        body = None

        if self.response_content_type is not None:
            headers = {
                "Content-Type": self.response_content_type
            }
            response = HttpResponseMessage(self.success_status, result, headers)
        else:
            if result is not None:
                body = convert_to_json(result)
            response = JsonHttpResponseMessage(self.success_status, body)

        message_bus.reply_message(message, response)

    def format_exception(self, message_bus, message, event, input_obj, exception):
        self.exception_formatter.format(message_bus, message, exception, 500)

    def format_parse_exception(self, message_bus, message, event, data, exception):
        self.exception_formatter.format(message_bus, message, exception, 400)



from datetime import date, datetime
from xml.etree import cElementTree as eTree

from model.order_detail import OrderDetail, AdditionalProperty, Tender, Item, \
    StateHistory
from model.services import OrderServiceModel
from iso8601 import parse_date
from typing import List


class OrderService(OrderServiceModel):

    def convert_mw_order_to_order_detail(self, order):
        # type: (eTree) -> OrderDetail

        order_detail = OrderDetail()

        order = order.find("Order") or order
        order_detail.order_id = int(order.get("orderId"))
        order_detail.sale_type = order.get("saleTypeDescr")
        order_detail.state = order.get("state")
        order_detail.total_amount = float(order.get("totalGross"))
        order_detail.total_tender = float(order.get("totalTender"))
        order_detail.due_amount = float(order.get("dueAmount"))
        order_detail.created_at = self._parse_timestamp(order.get("createdAt"))
        order_detail.state_history = self._build_state_history(order)
        order_detail.tenders = self._build_tenders(order)
        order_detail.additional_properties = self._build_additional_properties(order)
        order_detail.items = self._build_items(order)

        return order_detail

    def _build_tenders(self, order):
        # type: (eTree) -> List[Tender]

        tenders_list = []
        tenders = order.findall(".//Tender")
        for tender in tenders:
            tender_object = Tender()

            tender_object.sequence_id = int(tender.get("tenderId"))
            tender_object.type_id = int(tender.get("tenderType"))
            tender_object.type_name = tender.get("tenderDescr")
            tender_object.timestamp = self._parse_timestamp(tender.get("timestamp"))
            tender_object.amount = float(tender.get("tenderAmount"))

            tenders_list.append(tender_object)

        return tenders_list

    def _build_state_history(self, order):
        # type: (eTree) -> List[StateHistory]

        state_history_list = []
        states = order.findall(".//State")
        for state in states:
            state_object = StateHistory()

            state_object.state_name = state.get("state")
            state_object.timestamp = self._parse_timestamp(state.get("timestamp"))

            state_history_list.append(state_object)

        return state_history_list

    @staticmethod
    def _build_items(order):
        # type: (eTree) -> List[Item]

        items_list = []
        items = order.findall(".//SaleLine")
        for item in items:
            item_object = Item()

            item_object.context = item.get("itemId")
            item_object.part_code = int(item.get("partCode"))
            item_object.level = int(item.get("level"))
            item_object.line_number = int(item.get("lineNumber"))
            item_object.product_name = item.get("productName")
            item_object.item_type = item.get("itemType")
            item_object.added_price = float(item.get("addedUnitPrice")) if item.get("addedUnitPrice") else None
            item_object.unit_price = float(item.get("unitPrice")) if item.get("unitPrice") else None
            item_object.quantity = int(float(item.get("qty")))
            item_object.default_quantity = int(item.get("defaultQty") or 0)

            items_list.append(item_object)

        return items_list

    @staticmethod
    def _build_additional_properties(order):
        # type: (eTree) -> List[AdditionalProperty]

        additional_properties_list = []
        additional_properties = order.findall(".//OrderProperty")
        for additional_property in additional_properties:
            key = additional_property.get("key")
            value = additional_property.get("value")
            additional_property_object = AdditionalProperty(key, value)
            additional_properties_list.append(additional_property_object)

        return additional_properties_list

    def _parse_timestamp(self, timestamp):
        # type: (str) -> str

        timestamp = parse_date(timestamp)
        converted_time = self._convert_to_utc(timestamp)

        return converted_time.isoformat()

    @staticmethod
    def _convert_to_utc(local_datetime):
        # type: (date) -> date

        time_diff = datetime.utcnow() - datetime.now()
        converted_time = local_datetime + time_diff

        return converted_time



import json

from typing import List, Dict

from mbcontextmessagehandler import MbContextMessageBus
from mwhelper import BaseRepository
from persistence import Connection

from model.error_type import PosValidationErrorsType
from model.exception import ValidationError
from model.payment import PaymentType
from model.services import PaymentServiceModel


class PaymentService(PaymentServiceModel, BaseRepository):

    def __init__(
            self,
            message_bus     # type: MbContextMessageBus
    ):
        # type: (...) -> None

        super(PaymentService, self).__init__(message_bus.mbcontext)
        self.message_bus = message_bus
        self.payment_types = None

    def get_payment_types(
        self
    ):
        # type: () -> List[PaymentType]

        def inner_func(conn):
            # type: (Connection) -> List[int]

            query = """
                SELECT
                    TT.TenderId AS TenderId,
                    TT.TenderDescr AS TenderDescr,
                    TT.ElectronicType AS ElectronicTypeId,
                    TET.TypeDescr AS ElectronicTypeDescr,
                    TT.ParentTenderId AS ParentTenderId,
                    TT.CustomAttr As CustomAttr
                FROM productdb.TenderType TT
                JOIN productdb.TenderElectronicType TET ON TET.TypeId=TT.ElectronicType;
            """
            cursor = conn.select(query)

            tenders = {}
            for row in cursor:
                tender_id = int(row.get_entry('TenderId'))
                parent_tender_id = row.get_entry('ParentTenderId')
                tenders[tender_id] = {
                    'id': tender_id,
                    'descr': row.get_entry('TenderDescr'),
                    'electronicType': row.get_entry('ElectronicTypeDescr'),
                    'electronicTypeId': int(row.get_entry('ElectronicTypeId')),
                    'parentId ': int(parent_tender_id) if parent_tender_id else None,
                    'customAttr': row.get_entry('CustomAttr')
                }

            return tenders

        if not self.payment_types:
            self.payment_types = self.format_payment_types(self.execute_with_connection(
                inner_func,
                service_name="Persistence"
            ))

        return self.payment_types

    def format_payment_types(
            self,
            tender_types    # type: List[Dict[int, str]]
    ):
        # type: (...) -> List[PaymentType]

        payment_types = []

        for index, tender in tender_types.items():
            if "customAttr" not in tender or index >= 90:
                continue

            custom_attr = json.loads(tender["customAttr"] or "{}")
            show = custom_attr["showInTotem"] if "showInTotem" in custom_attr else False
            if not show:
                continue

            image_path = custom_attr["imagePath"] if "imagePath" in custom_attr else None
            payment_type = PaymentType(
                name=tender["descr"],
                payment_type=index,
                image_path=image_path
            )

            payment_types.append(payment_type)

        if not payment_types:
            raise ValidationError(PosValidationErrorsType.PaymentTypesToShowNotFound)

        return sorted(payment_types, key=lambda p: p.payment_type)



from logging import Logger

from production.box import ProductionBox
from production.model import Item
from typing import Dict, Optional, Union, List

from ._OrderChangerProductionBox import OrderChangerProductionBox


class CollapseSameItemsBox(OrderChangerProductionBox):
    def __init__(self, name, sons, include_seats, logger=None):
        # type: (str, Optional[Union[List[ProductionBox]], ProductionBox], bool, Logger) -> None

        super(CollapseSameItemsBox, self).__init__(name, sons, logger)

        self.include_seats = include_seats

    def change_order(self, order):
        new_items = []
        collapsed_items = {}  # type: Dict[str, Item]
        for item in order.items:
            key = self.build_item_key(item)
            if key not in collapsed_items:
                collapsed_items[key] = item
                if not hasattr(item, 'joined_items'):
                    item.joined_items = []
                new_items.append(item)
            else:
                collapsed_items[key].qty += item.qty
                collapsed_items[key].multiplied_qty += item.multiplied_qty
                collapsed_items[key].joined_items.append(item)

        order.items = new_items
        return order

    def build_item_key(self, item, key="", is_parent_product=False):
        # type: (Item, str, bool) -> str

        key += str(item.part_code)
        if self.include_seats:
            key += item.seat

        if len(item.comments) > 0:
            for comment in item.comments.values():
                key += comment.text

        if is_parent_product:
            key += str(item.qty)

        for tag in item.get_tags():
            key += tag

        for son in item.items:
            is_parent_product = item.item_type.upper() == "PRODUCT"
            key += self.build_item_key(son, key, is_parent_product=is_parent_product)

        return key



# -*- coding: utf-8 -*-
"""
    werkzeug.security
    ~~~~~~~~~~~~~~~~~

    Security related helpers such as secure password hashing tools.

    :copyright: 2007 Pallets
    :license: BSD-3-Clause
"""
import codecs
import hashlib
import hmac
import os
import posixpath
from random import SystemRandom
from struct import Struct

from ._compat import izip
from ._compat import PY2
from ._compat import range_type
from ._compat import text_type
from ._compat import to_bytes
from ._compat import to_native

SALT_CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
DEFAULT_PBKDF2_ITERATIONS = 150000

_pack_int = Struct(">I").pack
_builtin_safe_str_cmp = getattr(hmac, "compare_digest", None)
_sys_rng = SystemRandom()
_os_alt_seps = list(
    sep for sep in [os.path.sep, os.path.altsep] if sep not in (None, "/")
)


def pbkdf2_hex(
    data, salt, iterations=DEFAULT_PBKDF2_ITERATIONS, keylen=None, hashfunc=None
):
    """Like :func:`pbkdf2_bin`, but returns a hex-encoded string.

    .. versionadded:: 0.9

    :param data: the data to derive.
    :param salt: the salt for the derivation.
    :param iterations: the number of iterations.
    :param keylen: the length of the resulting key.  If not provided,
                   the digest size will be used.
    :param hashfunc: the hash function to use.  This can either be the
                     string name of a known hash function, or a function
                     from the hashlib module.  Defaults to sha256.
    """
    rv = pbkdf2_bin(data, salt, iterations, keylen, hashfunc)
    return to_native(codecs.encode(rv, "hex_codec"))


def pbkdf2_bin(
    data, salt, iterations=DEFAULT_PBKDF2_ITERATIONS, keylen=None, hashfunc=None
):
    """Returns a binary digest for the PBKDF2 hash algorithm of `data`
    with the given `salt`. It iterates `iterations` times and produces a
    key of `keylen` bytes. By default, SHA-256 is used as hash function;
    a different hashlib `hashfunc` can be provided.

    .. versionadded:: 0.9

    :param data: the data to derive.
    :param salt: the salt for the derivation.
    :param iterations: the number of iterations.
    :param keylen: the length of the resulting key.  If not provided
                   the digest size will be used.
    :param hashfunc: the hash function to use.  This can either be the
                     string name of a known hash function or a function
                     from the hashlib module.  Defaults to sha256.
    """
    if not hashfunc:
        hashfunc = "sha256"

    data = to_bytes(data)
    salt = to_bytes(salt)

    if callable(hashfunc):
        _test_hash = hashfunc()
        hash_name = getattr(_test_hash, "name", None)
    else:
        hash_name = hashfunc
    return hashlib.pbkdf2_hmac(hash_name, data, salt, iterations, keylen)


def safe_str_cmp(a, b):
    """This function compares strings in somewhat constant time.  This
    requires that the length of at least one string is known in advance.

    Returns `True` if the two strings are equal, or `False` if they are not.

    .. versionadded:: 0.7
    """
    if isinstance(a, text_type):
        a = a.encode("utf-8")
    if isinstance(b, text_type):
        b = b.encode("utf-8")

    if _builtin_safe_str_cmp is not None:
        return _builtin_safe_str_cmp(a, b)

    if len(a) != len(b):
        return False

    rv = 0
    if PY2:
        for x, y in izip(a, b):
            rv |= ord(x) ^ ord(y)
    else:
        for x, y in izip(a, b):
            rv |= x ^ y

    return rv == 0


def gen_salt(length):
    """Generate a random string of SALT_CHARS with specified ``length``."""
    if length <= 0:
        raise ValueError("Salt length must be positive")
    return "".join(_sys_rng.choice(SALT_CHARS) for _ in range_type(length))


def _hash_internal(method, salt, password):
    """Internal password hash helper.  Supports plaintext without salt,
    unsalted and salted passwords.  In case salted passwords are used
    hmac is used.
    """
    if method == "plain":
        return password, method

    if isinstance(password, text_type):
        password = password.encode("utf-8")

    if method.startswith("pbkdf2:"):
        args = method[7:].split(":")
        if len(args) not in (1, 2):
            raise ValueError("Invalid number of arguments for PBKDF2")
        method = args.pop(0)
        iterations = args and int(args[0] or 0) or DEFAULT_PBKDF2_ITERATIONS
        is_pbkdf2 = True
        actual_method = "pbkdf2:%s:%d" % (method, iterations)
    else:
        is_pbkdf2 = False
        actual_method = method

    if is_pbkdf2:
        if not salt:
            raise ValueError("Salt is required for PBKDF2")
        rv = pbkdf2_hex(password, salt, iterations, hashfunc=method)
    elif salt:
        if isinstance(salt, text_type):
            salt = salt.encode("utf-8")
        mac = _create_mac(salt, password, method)
        rv = mac.hexdigest()
    else:
        rv = hashlib.new(method, password).hexdigest()
    return rv, actual_method


def _create_mac(key, msg, method):
    if callable(method):
        return hmac.HMAC(key, msg, method)

    def hashfunc(d=b""):
        return hashlib.new(method, d)

    # Python 2.7 used ``hasattr(digestmod, '__call__')``
    # to detect if hashfunc is callable
    hashfunc.__call__ = hashfunc
    return hmac.HMAC(key, msg, hashfunc)


def generate_password_hash(password, method="pbkdf2:sha256", salt_length=8):
    """Hash a password with the given method and salt with a string of
    the given length. The format of the string returned includes the method
    that was used so that :func:`check_password_hash` can check the hash.

    The format for the hashed string looks like this::

        method$salt$hash

    This method can **not** generate unsalted passwords but it is possible
    to set param method='plain' in order to enforce plaintext passwords.
    If a salt is used, hmac is used internally to salt the password.

    If PBKDF2 is wanted it can be enabled by setting the method to
    ``pbkdf2:method:iterations`` where iterations is optional::

        pbkdf2:sha256:80000$salt$hash
        pbkdf2:sha256$salt$hash

    :param password: the password to hash.
    :param method: the hash method to use (one that hashlib supports). Can
                   optionally be in the format ``pbkdf2:<method>[:iterations]``
                   to enable PBKDF2.
    :param salt_length: the length of the salt in letters.
    """
    salt = gen_salt(salt_length) if method != "plain" else ""
    h, actual_method = _hash_internal(method, salt, password)
    return "%s$%s$%s" % (actual_method, salt, h)


def check_password_hash(pwhash, password):
    """check a password against a given salted and hashed password value.
    In order to support unsalted legacy passwords this method supports
    plain text passwords, md5 and sha1 hashes (both salted and unsalted).

    Returns `True` if the password matched, `False` otherwise.

    :param pwhash: a hashed string like returned by
                   :func:`generate_password_hash`.
    :param password: the plaintext password to compare against the hash.
    """
    if pwhash.count("$") < 2:
        return False
    method, salt, hashval = pwhash.split("$", 2)
    return safe_str_cmp(_hash_internal(method, salt, password)[0], hashval)


def safe_join(directory, *pathnames):
    """Safely join zero or more untrusted path components to a base
    directory to avoid escaping the base directory.

    :param directory: The trusted base directory.
    :param pathnames: The untrusted path components relative to the
        base directory.
    :return: A safe path, otherwise ``None``.
    """
    parts = [directory]

    for filename in pathnames:
        if filename != "":
            filename = posixpath.normpath(filename)

        if (
            any(sep in filename for sep in _os_alt_seps)
            or os.path.isabs(filename)
            or filename == ".."
            or filename.startswith("../")
        ):
            return None

        parts.append(filename)

    return posixpath.join(*parts)



from messageprocessor import MessageProcessorCallback


class DbApiTransactionCallback(MessageProcessorCallback):

    def __init__(self):
        self.conn = None

    def update_conn(self, conn):
        self.conn = conn

    def process_begun(self, event_name, request_id, data):
        return

    def input_parsed(self, event_name, request_id, input_obj):
        return

    def business_called(self, event_name, request_id, input_obj, output_obj):
        if self.conn:
            self.conn.commit()

    def process_finished(self, event_name, request_id, input_obj, output_obj):
        return

    def parse_exception(self, event_name, request_id, data, exception):
        self._rollback()

    def unhandled_exception(self, event_name, request_id, input_obj, exception):
        self._rollback()

    def _rollback(self):
        if self.conn:
            self.conn.rollback()



import json


class DeliveryOrder(object):
    def __init__(self, order_json):
        order = json.loads(order_json)

        self.store_name = self._get_store_name(order)
        self.order_number = order.get("shortReference", "0")
        self.order_date = order.get("createAt", "2099-01-01T00:00:00.00000Z")
        self.items = order.get("items", [])
        self.total_price = order.get("totalPrice", "0.00")
        self.sub_total_price = order.get("subTotal", "0.00")
        self.delivery_fee = order.get("deliveryFee", "0.00")
        self.observation = self._get_custom_property(order, "observation")
        self.observations = order.get("observations", "")
        self.tenders = order.get("tenders", [])
        self.customer_name = self._get_custom_property(order, "customer_name")
        self.customer_doc = self._get_custom_property(order, "customer_doc")
        self.customer_phone = self._get_custom_property(order, "customer_phone")
        self.address = ""
        self.address_neighborhood = ""
        self.city = ""
        self.address_complement = ""
        self.postal_code = ""
        self.vouchers = {}
        self.partner = order.get("partner", "")

        if "pickup" in order and "address" in order["pickup"]:
            self.address = order["pickup"]["address"].get("formattedAddress", "")
            self.address_neighborhood = order.get("pickup").get("address").get("neighborhood", "")
            self.city = order.get("pickup").get("address").get("city", "")
            self.address_complement = order.get("pickup").get("address").get("complement", "")
            self.postal_code = order.get("pickup").get("address").get("postalCode", "")

        self._create_vouchers(order)

    def _create_vouchers(self, order):
        self.vouchers["ORDER"] = 0
        self.vouchers["TOTAL_ITEMS"] = 0
        self.vouchers["DELIVERY_FEE"] = 0

        if "vouchers" in order and "store" in order["vouchers"]:
            self._get_order_vouchers(order)
            self._get_total_items_vouchers(order)
            self._get_delivery_fee_vouchers(order)

    def _get_order_vouchers(self, order):
        if "order" in order["vouchers"]["store"]:
            self.vouchers["ORDER"] += order["vouchers"]["store"]["order"]

        if "order" in order["vouchers"]["partner"]:
            self.vouchers["ORDER"] += order["vouchers"]["partner"]["order"]

    def _get_total_items_vouchers(self, order):
        if "total_items" in order["vouchers"]["store"]:
            self.vouchers["TOTAL_ITEMS"] += order["vouchers"]["store"]["total_items"]

        if "total_items" in order["vouchers"]["partner"]:
            self.vouchers["TOTAL_ITEMS"] += order["vouchers"]["partner"]["total_items"]

    def _get_delivery_fee_vouchers(self, order):
        if "delivery_fee" in order["vouchers"]["store"]:
            self.vouchers["DELIVERY_FEE"] += order["vouchers"]["store"]["delivery_fee"]

        if "delivery_fee" in order["vouchers"]["partner"]:
            self.vouchers["DELIVERY_FEE"] += order["vouchers"]["partner"]["delivery_fee"]

    @staticmethod
    def _get_custom_property(order, key):
        if "custom_properties" in order:
            value = [x for x in order["custom_properties"] if x["key"] == key]
            if len(value) > 0:
                return value[0]["value"]

        return ""

    @staticmethod
    def _get_store_name(order):
        store_name = order["partner"]
        if "brand" in order and order["brand"] is not None:
            store_name += " - " + order["brand"]

        return store_name



import logging
from datetime import datetime, timedelta
from threading import Lock

from sysactions import action
from application.interactor import NDiscountInteractor
from typing import List, Dict
from actions.coupon.get_coupon import is_coupon_component_connected

discounts_cache = []
discounts_available_cache = False
coupons_available_cache = False
last_cache_time = None
lock = Lock()
logger = logging.getLogger("PosActions")


@action
def is_discounts_available(pos_id, seconds_to_renew_cache=3600):
    global discounts_available_cache
    global coupons_available_cache

    get_visible_discounts(pos_id, seconds_to_renew_cache=seconds_to_renew_cache)
    if len(discounts_cache) > 0:
        discounts_available_cache = True
    else:
        discounts_available_cache = False

    return str(discounts_available_cache).lower()


def get_visible_discounts(pos_id, seconds_to_renew_cache=3600):
    global last_cache_time
    global discounts_cache
    global discounts_available_cache

    with lock:
        now = datetime.now()
        seconds_to_renew = _get_seconds_to_renew(seconds_to_renew_cache)
        component_connected = is_coupon_component_connected(pos_id)
        if not discounts_cache or _reached_seconds_to_renew_cache(now, seconds_to_renew):
            if not component_connected:
                discounts_cache = _get_static_discounts()
            else:
                discounts_cache = _get_discounts_from_database(pos_id, True)
            last_cache_time = now
        return discounts_cache, not component_connected


def _get_discounts_from_database(pos_id, only_visible):
    # type: (int, bool) -> List[Dict[str, any]]

    discount_interactor = NDiscountInteractor()
    coupons = discount_interactor.get_benefits(pos_id, only_visible)

    discounts = []
    for item in coupons:
        benefit_id = str(item.get("benefitId", ""))
        coupon = item.get("benefitJson", {})
        name = coupon.get("userDescription", "")
        discounts.extend([{"coupon": coupon,
                           "index": coupon.get("priority", 0),
                           "name": name,
                           "benefit_id": benefit_id}]
                         )
    discounts = sorted(discounts, key=lambda element: element["index"])
    for index, d in enumerate(discounts):
        d["index"] = index + 1

    return discounts


def _get_seconds_to_renew(seconds_to_renew_cache):
    seconds_to_renew = 180
    try:
        seconds_to_renew = int(seconds_to_renew_cache)
    except TypeError as _:
        pass
    return seconds_to_renew


def _reached_seconds_to_renew_cache(now, seconds_to_renew_cache):
    return not last_cache_time or now > last_cache_time + timedelta(seconds=seconds_to_renew_cache)


def _get_static_discounts():
    coupons = [{"benefit_id": 1, "name": 'Valor', "index": 1}, {"benefit_id": 2, "name": 'Porcentagem', "index": 2}]
    return coupons



from logging import Logger

from model.config import Config
from model.electronic_message import ElectronicMessage
from model.error_type import PosValidationErrorsType
from model.exception import ValidationError
from model.repository import PosRepositoryModel
from interactor.route.electronic_payment.send_responses.dto import SendResponsesRequest
from validation_util import validate_api_key, check_if_business_day_is_opened, \
    check_if_operator_is_logged, check_if_order_is_totaled
from typing import List, Any


class SendResponsesInteractor(object):
    def __init__(self, logger, config, pos_repository, fiscal_service, electronic_messages):
        # type: (Logger, Config, PosRepositoryModel, Any, List[ElectronicMessage]) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.fiscal_service = fiscal_service
        self.electronic_messages = electronic_messages

    def execute(self, request):
        # type: (SendResponsesRequest) -> None

        pos_id = validate_api_key(request, self.config.api_key_mapping)
        if not pos_id:
            raise ValidationError(PosValidationErrorsType.PosIdNotFound)

        model = self.pos_repository.get_model(pos_id)

        check_if_business_day_is_opened(self.pos_repository, model)
        check_if_operator_is_logged(self.pos_repository, model)
        check_if_order_is_totaled(self.pos_repository, model)

        self.fiscal_service.send_electronic_payment_message_response(model, request)



from httpmessageprocessorutil.automapped import auto_map


@auto_map({"system_api_key": {"name": "X-Api-Key", "location": "header", "type": str}})
class GetRupturedProductsRequest(object):
    def __init__(self, system_api_key):
        # type: (str) -> None

        self.system_api_key = system_api_key



from datetime import datetime

from model.electronic_message import ElectronicMessage
from model.event import ListenedEvents
from typing import Any

from interactor.event import ElectronicPaymentEventInteractor
from messageprocessorutil import UuidMessageProcessor


class ElectronicPaymentShowQuestionProcessor(UuidMessageProcessor):

    def __init__(self, interactor):
        # type: (ElectronicPaymentEventInteractor) -> None

        super(ElectronicPaymentShowQuestionProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(self, data, event_type):
        # type: (Any, str) -> ElectronicMessage

        pos_id, message, title, buttons = data.split("\\0")

        electronic_message = ElectronicMessage(
            pos_id=pos_id,
            event_type=ListenedEvents.ELECTRONIC_PAYMENT_SHOW_QUESTION,
            title=title,
            timestamp=datetime.now().isoformat(),
            buttons=buttons,
            message_data=message
        )
        return electronic_message

    def call_business(self, electronic_message):
        # type: (ElectronicMessage) -> None

        self.interactor.queue_message(electronic_message)

    def format_response(self, message_bus, message, event, input_obj, result):
        pass

    def format_exception(self, message_bus, message, event, input_obj, exception):
        pass

    def format_parse_exception(self, message_bus, message, event, data, exception):
        pass



import base64
from datetime import datetime, timedelta

from actions.eft import cancel_eft_transactions
from old_helper import convert_from_utf_to_localtime

import sysactions
from custom_sysactions import user_authorized_to_execute_action
from utils import get_void_reason_id, void_order, void_fiscal_order
from .. import logger


@sysactions.action
@user_authorized_to_execute_action
def do_void_order(
    pos_id,
    void_reason=None,
    fiscal_void='false',
    order_id="",
    show_cancel_message=True,
    is_mobile=False
):
    try:
        from posactions import get_printer_receipt
        model = sysactions.get_model(pos_id)
        originator_pos_id = pos_id

        if fiscal_void.lower() == 'true':
            order_id, originator_pos_id = _void_fiscal_order(pos_id, model)
        else:
            sysactions.check_current_order(pos_id, model)

        void_reason = get_void_reason_id(pos_id, void_reason)
        if not void_reason:
            return True

        void_order(originator_pos_id, order_id=order_id, void_reason=void_reason)
        from posactions import is_table_service
        is_table_service = is_table_service(model)
        is_mobile = str(is_mobile).lower() == 'true'
        cancel_eft_transactions(
            pos_id,
            model,
            order_id,
            tender=None,
            is_table_service=is_table_service,
            is_mobile=is_mobile
        )

        if not is_mobile:
            printer_receipt = get_printer_receipt()
            printer_receipt.print_tef_receipts(pos_id, cancel_receipt=True)

        if show_cancel_message:
            sysactions.show_messagebox(pos_id, "$ORDER_VOIDED_WITH_SUCCESS")

        return True
    except sysactions.StopAction as _:
        return False
    except BaseException as _:
        logger.exception("Error voiding order {}".format(str(order_id) if order_id else 'undefined'))
        if show_cancel_message:
            sysactions.show_messagebox(pos_id, "$ERROR_CANCEL_ORDER")

        return False


def _void_fiscal_order(pos_id, model):
    order_id = ""
    posot = sysactions.get_posot(model)
    limit_datetime = datetime.now() - timedelta(minutes=30)
    orders = posot.listOrders(state="PAID", since=limit_datetime.strftime("%Y-%m-%dT%H:%M:%S"))
    order_options = []
    baseurl = "/mwapp/services/PosController/POS{}/?token=TK_POS_EXECUTEACTION&format=2&isBase64=true&payload=" \
        .format(pos_id)
    order_descr = "{} - {} POS{:02} - {:05} - R${:.2f}"
    for order in orders:
        if 'custom_properties' in order and 'FISCALIZATION_DATE' in order['custom_properties']:
            fiscalization_prop = order['custom_properties']['FISCALIZATION_DATE']
            fiscalization_date = convert_from_utf_to_localtime(datetime.strptime(fiscalization_prop, "%Y-%m-%dT%H:%M:%S"))
            if fiscalization_date.replace(tzinfo=None) > limit_datetime:
                descr = order_descr.format(
                    fiscalization_date.strftime('%H:%M'),
                    order['podType'],
                    int(order['originatorId'][3:]),
                    int(order['orderId']),
                    float(order['totalGross'])
                )
                payload = base64.b64encode("\0".join(("doOrderPicture", pos_id, order['orderId'])))
                url = baseurl + payload
                order_options.append((order['orderId'], descr, url))
    if len(order_options) > 0:
        selected = sysactions.show_order_preview(pos_id,
                                                 order_options,
                                                 "$SELECT_AN_ORDER_TO_CANCEL",
                                                 buttons="CANCEL|OK",
                                                 onthefly=True)
        if (selected is None) or (selected[0] == "0"):
            raise sysactions.StopAction()
        order_id = selected[1]
    if order_id:
        originator_pos_id = int(next(iter([x for x in orders if x['orderId'] == order_id]))['originatorId'][3:])
        void_fiscal_order(originator_pos_id, order_id)
    else:
        sysactions.show_messagebox(pos_id, "$NO_ORDER_FOUND_TO_CANCEL")
        raise sysactions.StopAction()
    return order_id, originator_pos_id



from logging import Logger
from typing import List

from interactor.route.payment.get_payment_types.dto import GetPaymentTypesRequest
from model.config import Config
from model.error_type import PosValidationErrorsType
from model.exception import ValidationError
from model.payment import PaymentType
from model.repository import PosRepositoryModel
from model.services import PaymentServiceModel
from validation_util import validate_api_key


class GetPaymentTypesInteractor(object):

    def __init__(
            self,
            logger,             # type: Logger
            config,             # type: Config
            pos_repository,     # type: PosRepositoryModel
            payment_service     # type: PaymentServiceModel
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.payment_service = payment_service

    def execute(
            self,
            request     # type: GetPaymentTypesRequest
    ):
        # type: (...) -> List[PaymentType]

        validate_api_key(request, self.config.api_key_mapping)

        tender_types = self.payment_service.get_payment_types()

        if not tender_types:
            raise ValidationError(PosValidationErrorsType.PaymentTypesNotFound)

        return tender_types



from helper import json_serialize
from messagebus import Message, DefaultToken, DataType
from messageprocessorutil import UuidMessageProcessor

from customer_services.interactor.get_customer.dto import GetCustomerRequest
from customer_services.model import Customer
from customer_services.interactor.get_customer import GetCustomerInteractor

from externals.util import LoggerUtil

logger = LoggerUtil.get_logger_name(__name__)


class GetCustomerProcessor(UuidMessageProcessor):
    def __init__(self, interactor):
        # type: (GetCustomerInteractor) -> None
        
        super(GetCustomerProcessor, self).__init__(event_name=None)
        
        self.interactor = interactor

    def parse_data(self, phone):
        # type: (str) -> GetCustomerRequest
        
        return GetCustomerRequest(phone)
    
    def call_business(self, request):
        # type: (GetCustomerRequest) -> Customer
        
        logger.info("Getting customer from phone {}".format(request.phone))
        customer = self.interactor.get_customer(request)
        
        return customer
    
    def format_response(self, message_bus, message, event, request, response):
        response = json_serialize(response)
        logger.info("Customer returned from phone {}: {}".format(request.phone, response))
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.param, response))
    
    def format_exception(self, message_bus, message, event, request, exception):
        exception_name = str(type(exception).__name__)
        logger.info("Error getting customer from phone {}: {}".format(request.phone, exception_name))
        reply_message = Message(DefaultToken.TK_SYS_NAK, DataType.param, exception_name)
        message_bus.reply_message(message, reply_message)
    
    def format_parse_exception(self, message_bus, message, event, data, exception):
        logger.info("Error getting parsing request: {}".format(str(exception)))
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



import re

from datetime import datetime

from production.box._OrderChangerProductionBox import OrderChangerProductionBox
from production.model import Item, ProductionOrder


class AllDoneBlinkBox(OrderChangerProductionBox):
    def __init__(self, name, sons, logger=None):
        super(AllDoneBlinkBox, self).__init__(name, sons, logger)
        self.done_tags = ["done", "no-cook", "dont-need-cook"]

    def change_order(self, order):
        # type: (ProductionOrder) -> ProductionOrder

        for item in order.items:
            if not self._verify_item_tags(item):
                return order

        order.tagged_timestamp = self._get_last_done_timestamp(order)
        return order

    def _verify_item_tags(self, item):
        # type: (Item) -> bool

        if item.item_type == "COMBO":
            for son in item.items:
                if not self._verify_item_tags(son):
                    return False
        else:
            if item.voided:
                return True

            if not self._has_done_tags(item):
                return False

        return True

    def _has_done_tags(self, item):
        # type: (Item) -> bool

        has_generic_done_tag = any(x in item.tags for x in self.done_tags)
        if not item.displayed_jit or has_generic_done_tag:
            return has_generic_done_tag

        item_jit_line = item.jit_lines
        for jit in item.displayed_jit:
            if jit not in item_jit_line:
                continue

            jit_index = 0
            blink_sequence_jit = re.split('[><$]', item_jit_line)
            for jit_elem in blink_sequence_jit:
                if jit in re.findall('(\\w+)', jit_elem):
                    break
                jit_index += 1

            previous_jit_sequences = blink_sequence_jit[:jit_index]
            if not previous_jit_sequences:
                return False

            previous_jit_string = re.split('[><$]', "+".join(previous_jit_sequences))[-1]
            previous_jit_sequences = re.findall('\\w+', previous_jit_string)
            for jit_elem in previous_jit_sequences:
                if not self._part_code_is_done_in_jit(jit_elem, item):
                    return False

        return True

    @staticmethod
    def _part_code_is_done_in_jit(previous_jit, item):
        # type: (str, Item) -> bool

        return any([k for k, v in item.tags.iteritems() if v == "done-{}".format(previous_jit.lower())])

    def _get_last_done_timestamp(self, order):
        last_done = None
        for item in order.items:
            item_done = item.get_last_tag_list_time(self.done_tags)
            if item_done is None:
                continue
            if last_done is None or last_done < item_done:
                last_done = item_done
        if last_done is not None:
            return last_done.isoformat()
        return None



class LoginSummaryHeaderDto(object):
    def __init__(self, pos_id, store_id, operator_id, business_date):
        # type: (int, int, int, unicode) -> ()
        self.pos_id = pos_id
        self.store_id = store_id
        self.operator_id = operator_id
        self.business_date = business_date



from enum import Enum


class ActionTypes(Enum):
    itemsToAdd = "itemsToAdd"
    discountPercentageOnTotal = "discountPercentageOnTotal"
    discountPercentageByItems = "discountPercentageByItems"
    discountPercentageByGroupsOnTotal = "discountPercentageByGroupsOnTotal"
    discountAmountByGroupsOnTotal = "discountAmountByGroupsOnTotal"
    discountAmountOnTotal = "discountAmountOnTotal"
    discountAmountByItems = "discountAmountByItems"
    paymentAmount = "paymentAmount"
    paymentPercentage = "paymentPercentage"
    paymentAmountOnItem = "paymentAmountOnItem"
    discountPercentageByItemsOnTotal = "discountPercentageByItemsOnTotal"
    discountAmountByItemsOnTotal = "discountAmountByItemsOnTotal"



from _SaleChallenge import SaleChallenge
from ._Goals import Goals
from ._StoreGoals import StoreGoals
from ._ItemGoals import ItemGoals
from ._OperatorGoals import OperatorGoals
from ._GoalsCache import GoalsCache
from _DailyGoalsException import InvalidRequestException, DailyGoalsExistsException, DailyGoalsNonExistsException, \
    InsertDailyGoalsException, UpdateDailyGoalsException, DeleteDailyGoalsException, NotCacheGoals, \
    GetSalesChallengeException, InsertSalesChallengeException, DeleteSalesChallengeException, \
    UpdateSalesChallengeException



# encoding: utf-8

import locale
import xml.etree.cElementTree as eTree
from datetime import datetime

import sysactions
from typing import List, Optional

from repository import OrderRepository, ProductRepository
from .model import OrderItem
from helper import remove_accents


class PMixReport(object):
    TYPE_REAL_DATE = "RealDate"
    TYPE_BUSINESS_PERIOD = "BusinessPeriod"
    TYPE_SESSION_ID = "SessionId"
    TYPE_XML = "Xml"

    def __init__(self, order_repository, product_repository, store_id, operator_id):
        # type: (OrderRepository, ProductRepository, str, str) -> None
        self.order_repository = order_repository
        self.product_repository = product_repository
        self.store_id = store_id
        self.operator_id = operator_id

        try:
            locale.setlocale(locale.LC_ALL, "portuguese_brazil")
        except:
            locale.setlocale(locale.LC_ALL, "pt_BR.UTF-8")

    def generate_pmix_report_by_real_date(self, pos_id, initial_date, end_date, report_model, operatorid="operatorid"):
        # type: (int, Optional[datetime], Optional[datetime], str, Optional[str]) -> str
        return self._generate_pmix_report(self.TYPE_REAL_DATE, pos_id, initial_date, end_date, report_model, operatorid)

    def generate_pmix_report_by_business_period(self, pos_id, initial_date, end_date, report_model, operatorid="operatorid"):
        # type: (int, Optional[datetime], Optional[datetime], str, Optional[str]) -> str
        return self._generate_pmix_report(self.TYPE_BUSINESS_PERIOD, pos_id, initial_date, end_date, report_model, operatorid)

    def _generate_pmix_report(self, report_type, pos_id, initial_date, end_date, report_model, operator_id=""):
        # type: (unicode, int, datetime, datetime, str, Optional[str]) -> str

        if report_type == self.TYPE_REAL_DATE:
            paid_orders = self.order_repository.get_paid_orders_by_real_date(initial_date, end_date, operator_id)
        else:
            paid_orders = self.order_repository.get_paid_orders_by_business_period(initial_date, end_date, operator_id)

        order_items = self.order_repository.get_orders_items(paid_orders)
        order_items = self.filter_order_items(order_items, report_model)

        count_pos_error = filter(lambda x: x == "error", order_items)
        count_pos_error = len(count_pos_error)

        order_items = filter(lambda x: x != "error", order_items)

        order_items = self.product_repository.get_order_items_names(order_items)

        grouped_order_items = {}
        for item in order_items:
            if item.pcode not in grouped_order_items.keys():
                grouped_order_items[item.pcode] = item
            else:
                grouped_order_items[item.pcode].quantity += item.quantity

        order_items = grouped_order_items.values()
        order_items = sorted(order_items, key=lambda order_item: order_item.pcode)

        header = self._build_header(pos_id, initial_date, end_date, report_type, operator_id, report_model, count_pos_error)
        body = self._build_body(order_items)

        report = header + body
        report_bytes = report.encode("utf-8")

        return report_bytes

    @staticmethod
    def filter_order_items(order_items, report_model):
        if report_model == "PRICED":
            return [x for x in order_items if x.unit_price != 0]
        elif report_model == "UNPRICED":
            return [x for x in order_items if x.unit_price == 0]
        else:
            return order_items

    def generate_pmix_order_items(self, initial_date, end_date):
        # type: (Optional[datetime], Optional[datetime]) -> List[OrderItem]
        _, order_items = self._get_order_items(initial_date, end_date, self.TYPE_BUSINESS_PERIOD)
        return order_items

    def _get_order_items(self, initial_date, end_date, report_type):
        if report_type == self.TYPE_REAL_DATE:
            paid_orders = self.order_repository.get_paid_orders_by_real_date(initial_date, end_date, self.operator_id)
        else:
            paid_orders = self.order_repository.get_paid_orders_by_business_period(initial_date, end_date, self.operator_id)
        order_items = self.order_repository.get_orders_items(paid_orders)
        count_pos_error = filter(lambda x: x == "error", order_items)
        count_pos_error = len(count_pos_error)
        order_items = filter(lambda x: x != "error", order_items)
        order_items = self.product_repository.get_order_items_names(order_items)
        grouped_order_items = {}
        for item in order_items:
            if item.pcode not in grouped_order_items.keys():
                grouped_order_items[item.pcode] = item
            else:
                grouped_order_items[item.pcode].quantity += item.quantity
        order_items = grouped_order_items.values()
        order_items = sorted(order_items, key=lambda order_item: order_item.pcode)
        return count_pos_error, order_items

    def _build_header(self, pos_id, initial_date, end_date, report_type, operator_id, report_model, count_pos_error=0):
        # type: (int, datetime, datetime, str, str, str, Optional[int]) -> unicode
        if operator_id not in ["0", "", "None"]:
            operator_name = eTree.XML(sysactions.get_user_information(operator_id)).find(".//user").get("LongName")[:20]
        else:
            operator_name = "Todos"
        pos_list = "Todos" if len(self.order_repository.pos_list) > 1 else self.order_repository.pos_list[0]

        model = sysactions.get_model(pos_id)

        header = u"======================================\n"
        header += u"     Mix de Produtos ({0})     \n".format("Data Fiscal" if report_type == self.TYPE_REAL_DATE else "Data Negocio")
        header += u" Loja.........: " + self.store_id + "\n"
        header += u" Data/hora....: " + datetime.now().strftime("%d/%m/%Y %H:%M:%S") + "\n"
        if report_type != self.TYPE_SESSION_ID:
            header += u" {0:.<13}: {1} - {2}\n".format("Periodo", initial_date.strftime("%d/%m/%y"), end_date.strftime("%d/%m/%y"))
        header += u" Operador.....: " + u"{0}\n".format(operator_name[:20])
        temp_pos_list = u" POS incluidos: " + u"{0}".format(pos_list)
        header += temp_pos_list[:36]
        if len(temp_pos_list) > 36:
            header += (u"\n                %s" % temp_pos_list[36:56])
        if len(temp_pos_list) > 62:
            header += (u"\n                %s" % temp_pos_list[56:76])
        header += u"\n"
        header += u" Modelo.......: {}\n".format(remove_accents(sysactions.translate_message(model, report_model)))
        if count_pos_error > 0:
            header += u" POS excluidos: " + u"{0}\n".format(count_pos_error)
        header += u"======================================\n"

        return header

    @staticmethod
    def _build_body(order_items):
        # type: (List[OrderItem]) -> unicode
        body = u"Codigo   Descricao                 Qtd\n"
        body += u"--------------------------------------\n"
        for order_item in sorted(order_items, key=lambda x: x.name):
            body += u"{0:0>10}  {1:<18} {2:>7}\n".format(order_item.pcode,
                                                        remove_accents(order_item.name)[:18],
                                                        _str_float(order_item.quantity))
        body += u"======================================\n"

        return body


def _str_float(float_num):
    return locale.format("%.2f", float_num, True).replace(",", ".")



from typing import Dict, Optional, Union


class HttpRequest(object):
    def __init__(self, url, method, query=None, headers=None, body=None):
        # type: (str, str, Optional[Dict[str, str]], Optional[Dict[str, str]], Optional[Union[bytes, str]]) -> None
        self.url = url
        self.method = method
        self.query = query if query is not None else {}
        self.headers = headers if headers is not None else {}
        self.body = body
        if isinstance(body, str):
            self.body = body.encode("utf-8")

    def __repr__(self):
        return 'HttpRequest(url = "{}", method = "{}", headers = "{}", body = "{}")'.format(self.url,
                                                                                            self.method,
                                                                                            self.headers,
                                                                                            self.body)

    def __str__(self):
        return self.__repr__()



from application.repository import AccountRepository
from systools import sys_log_exception

from .. import mb_context

account_repository = AccountRepository(mb_context)


def get_drawer_initial_amount(pos_id, session):
    try:
        return account_repository.get_initial_amount(pos_id, session)
    except BaseException as _:
        sys_log_exception("Error in consult initial amount")



from unittest import TestCase
import simplejson as json

from decimal import Decimal

class AlternateInt(int):
    def __repr__(self):
        return 'invalid json'
    __str__ = __repr__


class AlternateFloat(float):
    def __repr__(self):
        return 'invalid json'
    __str__ = __repr__


# class AlternateDecimal(Decimal):
#     def __repr__(self):
#         return 'invalid json'


class TestSubclass(TestCase):
    def test_int(self):
        self.assertEqual(json.dumps(AlternateInt(1)), '1')
        self.assertEqual(json.dumps(AlternateInt(-1)), '-1')
        self.assertEqual(json.loads(json.dumps({AlternateInt(1): 1})), {'1': 1})

    def test_float(self):
        self.assertEqual(json.dumps(AlternateFloat(1.0)), '1.0')
        self.assertEqual(json.dumps(AlternateFloat(-1.0)), '-1.0')
        self.assertEqual(json.loads(json.dumps({AlternateFloat(1.0): 1})), {'1.0': 1})

    # NOTE: Decimal subclasses are not supported as-is
    # def test_decimal(self):
    #     self.assertEqual(json.dumps(AlternateDecimal('1.0')), '1.0')
    #     self.assertEqual(json.dumps(AlternateDecimal('-1.0')), '-1.0')



# coding=utf-8
import base64

from sysdefs import eft
from modelcommon.model import SitefResponse
from typing import (
    Any,
    Union
)

from model import (
    CancelPaymentOrderData,
    ConfirmFinishPaymentData,
    EftData,
    PaymentResponseData
)
from paymentprocessor import PaymentProcessor
from util import get_transaction_identifier


class FakePaymentProcessor(PaymentProcessor):
    def connection_status(self, test_connection_data):
        return True

    def add_payment(self, start_payment_data):
        pos_id = start_payment_data.pos_id
        session_id = start_payment_data.session_id
        transaction_identifier = get_transaction_identifier(
            order_id=start_payment_data.order_id,
            tender_id=start_payment_data.electronic_type
        )
        self.register_eft_activity(EftData(pos_id, session_id, eft.EFT_CREDITSALE, 1,
                                           start_payment_data.amount, orderid=start_payment_data.order_id))
        self.dialog_manager.close_dialogs(pos_id)
        return PaymentResponseData(
            media="",
            result="0",
            card_number="4111XXXXXXXXXXXX",
            auth_code="1302",
            approved_amount=start_payment_data.amount,
            id_auth="08496898000142",
            cnpj_auth="68536379000192",
            bandeira="4",
            receipt_merchant=base64.b64encode("Recibo de teste.\n MERCHANT\nSitef FAKE"),
            receipt_customer=base64.b64encode("Recibo de teste.\n CUSTOMER\nSitef FAKE"),
            expiration_date="05/2029",
            nsu="1234",
            transaction_processor="00051",
            last_digits="1111",
            sat_acquirer="012",
            nfce_brand="01",
            original_nfce_brand="01",
            nsu_host="4321",
            electronic_type=start_payment_data.electronic_type,
            transaction_identifier=transaction_identifier
        )

    def cancel_payment_order(
        self,
        cancel_payment_order_data   # type: CancelPaymentOrderData
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        response = SitefResponse()
        response.nsu = "123"
        response.nsu_host = "321"
        response.nsu_origin = "555"
        response.transaction_value = "10"
        response.customer_receipt = base64.b64encode("Recibo de teste CANCELAMENTO.\n CUSTOMER\nSitef FAKE")
        response.merchant_receipt = base64.b64encode("Recibo de teste CANCELAMENTO.\n MERCHANT\nSitef FAKE")
        response.transaction_identifier = get_transaction_identifier(
            order_id=cancel_payment_order_data.order_id,
            tender_id=cancel_payment_order_data.tender_id
        )
        return response

    def cancel_payment(self):
        return ""

    def finish_payment(self, finish_payment_data):
        return True

    def confirm_finish_payment(
        self,
        confirm_finish_payment_data     # type: ConfirmFinishPaymentData
    ):
        # type: (...) -> None

        pass

    def get_name(self):
        return "FakeElectronicPayment"
